/**
 * @license
 * ReduxedChromeStorage v2.4.0
 * https://github.com/hindmost/reduxed-chrome-storage
 * Copyright (c) Savr Goryaev aka hindmost
 *
 * This source code is licensed under the MIT license
 * https://github.com/hindmost/reduxed-chrome-storage/blob/master/LICENSE
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.reduxedChromeStorage = factory());
}(this, (function () { 'use strict';

  function cloneDeep(o) {
      if (o == null || typeof o !== 'object')
          { return o; }
      if (Array.isArray(o))
          { return o.map(function (v) { return cloneDeep(v); }); }
      return Object.keys(o).reduce(function (acc, key) {
          return acc[key] = cloneDeep(o[key]), acc;
      }, {});
  }
  function isEqual(a, b) {
      if (a === b)
          { return true; }
      if (a == null || typeof a !== 'object' || b == null || typeof b !== 'object')
          { return false; }
      var keysA = Object.keys(a), keysB = Object.keys(b);
      if (keysA.length !== keysB.length)
          { return false; }
      for (var i = 0, list = keysA; i < list.length; i += 1) {
          var key = list[i];

          if (keysB.indexOf(key) <= -1 || !isEqual(a[key], b[key]))
              { return false; }
      }
      return true;
  }
  function mergeOrReplace(a, b) {
      if (Array.isArray(b))
          { return cloneDeep(b); }
      if (typeof a !== 'object' || Array.isArray(a) || typeof b !== 'object')
          { return typeof b !== 'undefined' ? b : a; }
      return Object.keys(a).concat(Object.keys(b).filter(function (key) { return !a[key]; })).reduce(function (acc, key) {
          return acc[key] = mergeOrReplace(a[key], b[key]), acc;
      }, {});
  }

  var ReduxedStorage = function ReduxedStorage(ref) {
      var createStore = ref.createStore;
      var reducer = ref.reducer;
      var storage = ref.storage;
      var bufferLife = ref.bufferLife;
      var initialState = ref.initialState;
      var enhancer = ref.enhancer;

      this.createStore = createStore;
      this.storage = storage;
      this.reducer = reducer;
      this.enhancer = enhancer;
      this.buffLife = bufferLife ? Math.min(Math.max(bufferLife, 0), 2000) : 100;
      this.state0 = initialState;
      this.state = null;
      this.buffStore = null;
      this.lastState = null;
      this.listeners = [];
      this.inited = false;
      this.dispatch = this.dispatch.bind(this);
      this.subscribe = this.subscribe.bind(this);
  };
  ReduxedStorage.prototype.init = function init () {
          var this$1$1 = this;

      if (this.inited)
          { return new Promise(function (resolve) {
              resolve(this$1$1);
          }); }
      var defaultState = this._createStore().getState();
      // subscribe for changes in chrome.storage
      this.storage.subscribe(function (data, oldData) {
          if (isEqual(data, this$1$1.state))
              { return; }
          this$1$1._setState(data);
          for (var i = 0, list = this$1$1.listeners; i < list.length; i += 1) {
              var fn = list[i];

                  fn(oldData);
          }
      });
      this.inited = true;
      // return a promise to be resolved when the last state (if any) is restored from chrome.storage
      return new Promise(function (resolve) {
          // try to restore the last state stored in chrome.storage, if any
          this$1$1.storage.load(function (storedState) {
              var state = storedState ?
                  mergeOrReplace(defaultState, storedState) : defaultState;
              if (this$1$1.state0) {
                  state = mergeOrReplace(state, this$1$1.state0);
              }
              this$1$1._setState(state);
              if (!isEqual(state, storedState)) {
                  this$1$1._send2Storage(state);
              }
              resolve(this$1$1);
          });
      });
  };
  ReduxedStorage.prototype.initFrom = function initFrom (state) {
      this._setState(state);
      this.inited = true;
      return this;
  };
  ReduxedStorage.prototype.uninit = function uninit () {
          var this$1$1 = this;

      return new Promise(function (resolve) {
          resolve(this$1$1);
      });
  };
  ReduxedStorage.prototype._createStore = function _createStore (initialState) {
      return this.createStore(this.reducer, initialState, this.enhancer);
  };
  ReduxedStorage.prototype._send2Storage = function _send2Storage (data) {
      this.storage.save(data);
  };
  ReduxedStorage.prototype._setState = function _setState (data) {
      if (data) {
          this.state = cloneDeep(data);
      }
  };
  ReduxedStorage.prototype.getState = function getState () {
      return this.state;
  };
  ReduxedStorage.prototype.subscribe = function subscribe (fn) {
          var this$1$1 = this;

      typeof fn === 'function' && this.listeners.push(fn);
      return function () {
          if (typeof fn === 'function') {
              this$1$1.listeners = this$1$1.listeners.filter(function (v) { return v !== fn; });
          }
      };
  };
  ReduxedStorage.prototype.dispatch = function dispatch (action) {
          var this$1$1 = this;

      if (!this.buffStore) {
          // this.buffStore is to be used with sync actions
          this.buffStore = this._createStore(this.state);
          // this.lastState is shared by both sync and async actions
          this.lastState = this.buffStore.getState();
          setTimeout(function () {
              this$1$1.buffStore = null;
          }, this.buffLife);
      }
      // lastStore, holding an extra reference to the last created store, is to be used with async actions (e.g. via Redux Thunk);
      // then when this.buffStore is reset to null this variable should still refer to the same store
      var lastStore = this.buffStore;
      // set up a one-time state change listener
      var unsubscribe = lastStore.subscribe(function () {
          // if this.buffStore is non-empty, use it for getting the current state,
          // otherwise an async action is implied, so use lastStore instead
          var store = this$1$1.buffStore || lastStore;
          var state = store && store.getState();
          // we need a state change to be effective, so the current state should differ from the last saved one
          if (isEqual(state, this$1$1.lastState))
              { return; }
          // send the current state to chrome.storage & update this.lastState
          this$1$1._send2Storage(state);
          this$1$1.lastState = state;
          // as we already catched the 1st effective state change, we don't need this listener and lastStore anymore,
          // so we unsubscribe the former and reset the latter in order to release the related resources
          unsubscribe();
          lastStore = null;
      });
      return lastStore.dispatch(action);
  };
  ReduxedStorage.prototype.replaceReducer = function replaceReducer (nextReducer) {
      if (typeof nextReducer === 'function') {
          this.reducer = nextReducer;
      }
      return this;
  };
  ReduxedStorage.prototype[Symbol.observable] = function () {
          var obj;

      var getState = this.getState;
      var subscribe = this.subscribe;
      return ( obj = {
          subscribe: function subscribe$1(observer) {
              if (typeof observer !== 'object' || observer === null) {
                  throw new TypeError('Expected the observer to be an object.');
              }
              function observeState() {
                  var observerAsObserver = observer;
                  observerAsObserver.next && observerAsObserver.next(getState());
              }
              observeState();
              var unsubscribe = subscribe(observeState);
              return { unsubscribe: unsubscribe };
          }
      }, obj[Symbol.observable] = function () {
              return this;
          }, obj );
  };

  var StorageAreaName;
  (function (StorageAreaName) {
      StorageAreaName["local"] = "local";
      StorageAreaName["sync"] = "sync";
  })(StorageAreaName || (StorageAreaName = {}));

  var WrappedStorage = function WrappedStorage(ref) {
      var namespace = ref.namespace;
      var area = ref.area;
      var key = ref.key;

      this.ns = namespace;
      this.areaName = area === StorageAreaName.sync ? StorageAreaName.sync :
          StorageAreaName.local;
      this.key = key || 'reduxed';
      this.keyLen = this.key.length;
      this.lastSize = 0;
      this.listeners = [];
      this.errorListeners = [];
  };
  WrappedStorage.prototype.init = function init () {
          var this$1$1 = this;

      // Setup internal (shared) listener for chrome.storage.onChanged
      this.ns.storage.onChanged.addListener(function (changes, area) {
          if (area !== this$1$1.areaName || !(this$1$1.key in changes))
              { return; }
          var ref = changes[this$1$1.key];
              var newValue = ref.newValue;
              var oldValue = ref.oldValue;
          if (!newValue)
              { return; }
          this$1$1.lastSize = JSON.stringify(newValue).length;
          // call external chrome.storage.onChanged listeners
          for (var i = 0, list = this$1$1.listeners; i < list.length; i += 1) {
              var fn = list[i];

                  fn(newValue, oldValue);
          }
      });
  };
  WrappedStorage.prototype.subscribe = function subscribe (fn) {
      typeof fn === 'function' && this.listeners.push(fn);
  };
  WrappedStorage.prototype.subscribeForError = function subscribeForError (fn) {
      typeof fn === 'function' && this.errorListeners.push(fn);
  };
  WrappedStorage.prototype.fireErrorListeners = function fireErrorListeners (message, exceeded) {
      for (var i = 0, list = this.errorListeners; i < list.length; i += 1) {
          var fn = list[i];

              fn(message, exceeded);
      }
  };
  WrappedStorage.prototype.getErrorMessage = function getErrorMessage () {
      if (!this.ns.runtime.lastError)
          { return; }
      var ref = this.ns.runtime.lastError;
          var message = ref.message;
      return message || '';
  };
  WrappedStorage.prototype.callbackOnLoad = function callbackOnLoad (data, callback) {
      data = !this.ns.runtime.lastError && data && data[this.key];
      if (!this.lastSize && data) {
          this.lastSize = JSON.stringify(data).length;
      }
      callback(data);
  };
  WrappedStorage.prototype.checkQuotaPerItem = function checkQuotaPerItem (msg, area, data) {
      var b = this.areaName === StorageAreaName.sync &&
          area && area.QUOTA_BYTES_PER_ITEM && data &&
          JSON.stringify(data).length + this.keyLen > area.QUOTA_BYTES_PER_ITEM;
      b && this.fireErrorListeners(msg);
      return b;
  };
  WrappedStorage.prototype.checkQuota = function checkQuota (msg, area, data, total) {
      var b = !this.ns.runtime.lastError &&
          area && area.QUOTA_BYTES && data && total > 0 && this.lastSize > 0 &&
          JSON.stringify(data).length - this.lastSize > area.QUOTA_BYTES - total;
      this.fireErrorListeners(msg, b);
  };

  var WrappedChromeStorage = /*@__PURE__*/(function (WrappedStorage) {
      function WrappedChromeStorage(ref) {
          var namespace = ref.namespace;
          var area = ref.area;
          var key = ref.key;

          WrappedStorage.call(this, { namespace: namespace, area: area, key: key });
          this.areaApi = this.ns.storage[this.areaName];
      }

      if ( WrappedStorage ) WrappedChromeStorage.__proto__ = WrappedStorage;
      WrappedChromeStorage.prototype = Object.create( WrappedStorage && WrappedStorage.prototype );
      WrappedChromeStorage.prototype.constructor = WrappedChromeStorage;
      WrappedChromeStorage.prototype.load = function load (fn) {
          var this$1$1 = this;

          typeof fn === 'function' &&
              this.areaApi.get(this.key, function (data) {
                  this$1$1.callbackOnLoad(data, fn);
              });
      };
      WrappedChromeStorage.prototype.save = function save (data) {
          var this$1$1 = this;
          var obj;

          this.areaApi.set(( obj = {}, obj[this.key] = data, obj ), function () {
              var message = this$1$1.getErrorMessage();
              typeof message !== 'undefined' &&
                  !this$1$1.checkQuotaPerItem(message, this$1$1.areaApi, data) &&
                  this$1$1.areaApi.getBytesInUse(null, function (total) {
                      this$1$1.checkQuota(message, this$1$1.areaApi, data, total);
                  });
          });
      };

      return WrappedChromeStorage;
  }(WrappedStorage));

  var WrappedBrowserStorage = /*@__PURE__*/(function (WrappedStorage) {
      function WrappedBrowserStorage(ref) {
          var namespace = ref.namespace;
          var area = ref.area;
          var key = ref.key;

          WrappedStorage.call(this, { namespace: namespace, area: area, key: key });
          this.areaApi = this.ns.storage[this.areaName];
      }

      if ( WrappedStorage ) WrappedBrowserStorage.__proto__ = WrappedStorage;
      WrappedBrowserStorage.prototype = Object.create( WrappedStorage && WrappedStorage.prototype );
      WrappedBrowserStorage.prototype.constructor = WrappedBrowserStorage;
      WrappedBrowserStorage.prototype.load = function load (fn) {
          var this$1$1 = this;

          typeof fn === 'function' &&
              this.areaApi.get(this.key).then(function (data) {
                  this$1$1.callbackOnLoad(data, fn);
              });
      };
      WrappedBrowserStorage.prototype.save = function save (data) {
          var this$1$1 = this;
          var obj;

          this.areaApi.set(( obj = {}, obj[this.key] = data, obj )).then(function () {
              var message = this$1$1.getErrorMessage();
              typeof message !== 'undefined' &&
                  !this$1$1.checkQuotaPerItem(message, this$1$1.areaApi, data) &&
                  this$1$1.areaApi.getBytesInUse(null).then(function (total) {
                      this$1$1.checkQuota(message, this$1$1.areaApi, data, total);
                  });
          });
      };

      return WrappedBrowserStorage;
  }(WrappedStorage));

  var Namespace;
  (function (Namespace) {
      Namespace["chrome"] = "chrome";
      Namespace["browser"] = "browser";
  })(Namespace || (Namespace = {}));
  /**
   * ReduxedChromeStorage creator factory.
   * Returns an async store creator that's supposed to replace
   * the original Redux's createStore function.
   * Unlike the original createStore() that immediately returns a store,
   * async store creator returns a Promise to be resolved
   * when the created store is ready
   * @param obj
   * @param obj.createStore the original Redux's createStore function.
   * The only mandatory property/option
   * @param obj.namespace string to identify the APIs namespace to be used,
   * either 'chrome' or 'browser'.
   * If this and the next two options are missing,
   * the chrome namespace is used by default
   * @param obj.chromeNs the chrome namespace within Manifest V2 extension.
   * If this option is supplied, the previous one is ignored
   * @param obj.browserNs the browser namespace within Firefox extension,
   * or the chrome namespace within Manifest V3 chrome extension.
   * One may pass the chrome namespace within Manifest V3 to make this library
   * use Promise-based APIs under the hood.
   * If this option is supplied, the previous two are ignored
   * @param obj.changeListener a function to be called whenever the state changes,
   * receives two parameters:
   * 1) a one-time store - container of the current state;
   * 2) the previous state.
   * This option only makes sense in Manifest V3 service workers
   * or event-driven background scripts.
   * If this option is supplied, the async store creator returned by the factory
   * is not supposed to be immediately used for store creation;
   * its only purpose in this case is to hold the arguments to be passed
   * to the original `createStore` upon a one-time store creation
   * @param obj.errorListener a function to be called whenever an error occurs
   * during chrome.storage update, receives two parameters:
   * 1) an error message defined by storage API;
   * 2) a boolean indicating if the limit for the used storage area is exceeded
   * @param obj.storageArea the name of chrome.storage area to be used,
   * either 'local' or 'sync', defaults to 'local'
   * @param obj.storageKey key under which the state will be stored/tracked
   * in chrome.storage, defaults to 'reduxed'
   * @param obj.bufferLife lifetime of the bulk actions buffer (in ms),
   * defaults to 100
   * @returns an async store creator to replace the original createStore function
   */
  function reduxedStorageCreatorFactory(ref) {
      var createStore = ref.createStore;
      var namespace = ref.namespace;
      var chromeNs = ref.chromeNs;
      var browserNs = ref.browserNs;
      var changeListener = ref.changeListener;
      var errorListener = ref.errorListener;
      var storageArea = ref.storageArea;
      var storageKey = ref.storageKey;
      var bufferLife = ref.bufferLife;

      if (typeof createStore !== 'function')
          { throw new Error("Missing 'createStore' property/option"); }
      var storage = browserNs || namespace === Namespace.browser ?
          new WrappedBrowserStorage({
              namespace: browserNs || browser, area: storageArea, key: storageKey
          }) :
          new WrappedChromeStorage({
              namespace: chromeNs || chrome, area: storageArea, key: storageKey
          });
      storage.init();
      typeof errorListener === 'function' &&
          storage.subscribeForError(errorListener);
      function asyncStoreCreator(reducer, initialState, enhancer) {
          if (typeof reducer !== 'function')
              { throw new Error("Missing 'reducer' parameter"); }
          if (typeof initialState === 'function' && typeof enhancer === 'function')
              { throw new Error("Multiple 'enhancer' parameters unallowed"); }
          if (typeof initialState === 'function' && typeof enhancer === 'undefined') {
              enhancer = initialState;
              initialState = undefined;
          }
          var store = new ReduxedStorage({
              createStore: createStore, storage: storage, bufferLife: bufferLife,
              reducer: reducer, initialState: initialState, enhancer: enhancer
          });
          if (typeof changeListener !== 'function')
              { return store.init(); }
          storage.subscribe(function (data, oldData) {
              changeListener(store.initFrom(data), oldData);
          });
          return store.uninit();
      }
      return asyncStoreCreator;
  }

  return reduxedStorageCreatorFactory;

})));
